import warnings
import inspect
import numpy as np
import pandas as pd
import scipy.special
from lmfit.model import Model
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.font_manager import FontProperties
from iminuit import Minuit

FWHM_SIG_RATIO = np.sqrt(8 * np.log(2))  # 2.35482
SQRT_TWO = np.sqrt(2)  # 1.414213562
COLORS = [
    matplotlib.colors.to_rgb(c)
    for c in ["C0", "C2", "C4", "C5", "C6", "C7", "C8", "C9"]
]


class FittingError(Exception):
    """Exception raised by Fitters."""

    pass


class FittingWarning(UserWarning):
    """Warning raised by Fitters."""


# -----------------------------------------------------------------------------
# Base functions
# TODO: Should these be replaced with lmfit.models???
# -----------------------------------------------------------------------------


def constant(x, c):
    return np.ones_like(x) * c


def line(x, m, b):
    return m * x + b


def gauss(x, amp, mu, sigma):
    return (
        amp
        / sigma
        / np.sqrt(2.0 * np.pi)
        * np.exp(-((x - mu) ** 2.0) / (2.0 * sigma ** 2.0))
    )


def erf(x, amp, mu, sigma):
    return amp * 0.5 * (1.0 - scipy.special.erf((x - mu) / (SQRT_TWO * sigma)))


def exp(x, amp, lam):
    return amp * np.exp(x / lam)


def gausserf(x, ampgauss, amperf, mu, sigma):
    return gauss(x, ampgauss, mu, sigma) + erf(x, amperf, mu, sigma)


def expgauss(x, amp=1, mu=0, sigma=1.0, gamma=1.0):
    gss = gamma * sigma * sigma
    arg1 = gamma * (mu + gss / 2.0 - x)
    arg2 = (mu + gss - x) / (SQRT_TWO * sigma)
    return amp * (gamma / 2) * np.exp(arg1) * scipy.special.erfc(arg2)


# -----------------------------------------------------------------------------
# Custom loss functions
# -----------------------------------------------------------------------------


def poisson_loss(y_eval, y_data):
    return np.sum(y_eval - scipy.special.xlogy(y_data, y_eval))


# -----------------------------------------------------------------------------
# Fitting models
# -----------------------------------------------------------------------------


# Helper functions for guessing
def _xy_right(y, x=None, num=4):
    """Compute mean x and y in the last `num` points of a dataset (x, y).

    Parameters
    ----------
    y : array-like
        Y-data
    x : array-like, optional
        X-data. If not specified, return len(y)/2.
    num : int, optional
        Number of points to include in the averaging; default 4.

    Returns
    -------
    (float, float)
        Tuple of (xmean, ymean).
    """
    if x is not None:
        return np.mean(x[-num:]), np.mean(y[-num:])
    return len(y) * 0.5, np.mean(y[-num:])


def _xy_left(y, x=None, num=4):
    """Compute mean x and y in the first `num` points of a dataset (x, y).

    Parameters
    ----------
    y : array-like
        Y-data
    x : array-like, optional
        X-data. If not specified, return len(y)/2.
    num : int, optional
        Number of points to include in the averaging; default 4.

    Returns
    -------
    (float, float)
        Tuple of (xmean, ymean).
    """
    if x is not None:
        return np.mean(x[:num]), np.mean(y[:num])
    return len(y) * 0.5, np.mean(y[:num])


def _is_count_like(y):
    """Determine whether data 'looks like' it could be counts.

    Parameters
    ----------
    y : array-like
        Y-data
    """
    if np.any(y < 0):
        return False
    if not np.allclose(np.rint(y).astype(int), y):
        return False
    return True


class ConstantModel(Model):
    def __init__(self, *args, **kwargs):
        super().__init__(constant, *args, **kwargs)
        # TODO: remove this min setting?
        self.set_param_hint(f"{self.prefix}c", min=0.0)

    def guess(self, y, x=None, dx=None, num=2):
        if dx is None:
            dx = np.ones_like(x)
        c = (y[-1] / dx[-1] + y[0] / dx[0]) * 0.5
        return [
            (f"{self.prefix}c", "value", c),
            (f"{self.prefix}c", "min", 0.0),
        ]


class LineModel(Model):
    def __init__(self, *args, **kwargs):
        super().__init__(line, *args, **kwargs)

    def guess(self, y, x=None, dx=None, num=2):
        if dx is None:
            dx = np.ones_like(x)
        m = (y[-1] / dx[-1] - y[0] / dx[0]) / (x[-1] - x[0])
        b = ((y[-1] / dx[-1] + y[0] / dx[0]) - m * (x[1] + x[0])) * 0.5
        return [
            (f"{self.prefix}m", "value", m),
            (f"{self.prefix}b", "value", b),
        ]


class GaussModel(Model):
    def __init__(self, *args, **kwargs):
        super().__init__(gauss, *args, **kwargs)
        self.set_param_hint(
            f"{self.prefix}fwhm",
            expr=f"{FWHM_SIG_RATIO} * {self.prefix}sigma",
        )

    def guess(self, y, x=None, dx=None, center_ratio=0.5, width_ratio=0.5):
        assert center_ratio < 1, "Center mask ratio cannot exceed 1: {}".format(
            center_ratio
        )
        assert width_ratio < 1, "Width mask ratio cannot exceed 1: {}".format(
            width_ratio
        )

        if x is None:
            x = np.arange(0, len(y))
        if dx is None:
            dx = np.ones_like(x)

        xspan = x[-1] - x[0]
        mu = x[0] + xspan * center_ratio
        msk = (x >= (mu - xspan * width_ratio)) & (x <= mu + xspan * width_ratio)

        # TODO: update this, minimizer creates NaN's if default sigma used (0)
        sigma = xspan * width_ratio / 10.0
        amp = np.max(y[msk]) * np.sqrt(2 * np.pi) * sigma  # new amplitude guess
        return [
            (f"{self.prefix}amp", "value", amp),
            (f"{self.prefix}amp", "min", 0.0),
            (f"{self.prefix}mu", "value", mu),
            (f"{self.prefix}mu", "min", x[0]),
            (f"{self.prefix}mu", "max", x[-1]),
            (f"{self.prefix}sigma", "value", sigma),
            (f"{self.prefix}sigma", "min", 0.0),
        ]


class ErfModel(Model):
    def __init__(self, *args, **kwargs):
        super().__init__(erf, *args, **kwargs)

    def guess(self, y, x=None, dx=None, center_ratio=0):
        xspan = x[-1] - x[0]
        mu = x[0] + xspan * center_ratio
        return [
            (f"{self.prefix}amp", "value", y[0] - y[-1]),
            (f"{self.prefix}mu", "value", mu),
            (f"{self.prefix}sigma", "expr", "gauss_sigma"),
        ]


class GaussErfModel(Model):
    def __init__(self, *args, **kwargs):
        super().__init__(gausserf, *args, **kwargs)
        self.set_param_hint(
            f"{self.prefix}fwhm",
            expr=f"{FWHM_SIG_RATIO} * {self.prefix}sigma",
        )

    def guess(
        self, y, x=None, dx=None, center_ratio=0.5, width_ratio=0.5, amp_ratio=0.9
    ):
        assert center_ratio < 1, "Center mask ratio cannot exceed 1: {}".format(
            center_ratio
        )
        assert width_ratio < 1, "Width mask ratio cannot exceed 1: {}".format(
            width_ratio
        )
        if x is None:
            x = np.arange(0, len(y))
        if dx is None:
            dx = np.ones_like(x)
        xspan = x[-1] - x[0]
        mu = x[0] + xspan * center_ratio
        msk = (x >= (mu - xspan * width_ratio)) & (x <= mu + xspan * width_ratio)

        sigma = xspan * width_ratio / 10.0
        amp = np.max(y[msk]) * np.sqrt(2 * np.pi) * sigma
        amp_gauss = amp * amp_ratio
        amp_erf = amp * (1.0 - amp_ratio) / dx[0] / (np.sqrt(2 * np.pi) * sigma)
        return [
            (f"{self.prefix}ampgauss", "value", amp_gauss),
            (f"{self.prefix}ampgauss", "min", 0.0),
            (f"{self.prefix}amperf", "value", amp_erf),
            (f"{self.prefix}amperf", "min", 0.0),
            (f"{self.prefix}mu", "value", mu),
            (f"{self.prefix}mu", "min", x[0]),
            (f"{self.prefix}mu", "max", x[-1]),
            (f"{self.prefix}sigma", "value", sigma),
            (f"{self.prefix}sigma", "min", 0.0),
        ]


class ExpModel(Model):
    def __init__(self, *args, **kwargs):
        super().__init__(exp, *args, **kwargs)
        self.set_param_hint(f"{self.prefix}amp", min=0.0)
        self.set_param_hint(f"{self.prefix}lam", max=0.0)

    def guess(self, y, x=None, dx=None, num=1):
        xl, yl = _xy_left(y, x=x, num=num)
        xr, yr = _xy_right(y, x=x, num=num)
        # TODO: update this hardcoded zero offset
        lam = (xl - xr) / np.log(yl / (yr + 0.0001))
        amp = yl / np.exp(xl / lam)
        return [
            (f"{self.prefix}lam", "value", lam),
            (f"{self.prefix}lam", "max", -1e-3),
            (f"{self.prefix}amp", "value", amp),
            (f"{self.prefix}amp", "min", 0.0),
        ]


class ExpGaussModel(Model):
    """A model of an Exponentially modified Gaussian distribution
    (see https://en.wikipedia.org/wiki/Exponentially_modified_Gaussian_distribution)
    """

    def __init__(self, *args, **kwargs):
        super().__init__(expgauss, **kwargs)
        self.set_param_hint(f"{self.prefix}sigma", min=0)
        self.set_param_hint(f"{self.prefix}gamma", min=0, max=1)
        # TODO: This is obviously wrong
        self.set_param_hint(
            f"{self.prefix}fwhm",
            expr=f"{FWHM_SIG_RATIO} * {self.prefix}sigma",
        )

    def guess(self, y, x=None, dx=None, center_ratio=0.5, width_ratio=0.5):
        assert center_ratio < 1, "Center mask ratio cannot exceed 1: {}".format(
            center_ratio
        )
        assert width_ratio < 1, "Width mask ratio cannot exceed 1: {}".format(
            width_ratio
        )
        if x is None:
            x = np.arange(0, len(y))
        if dx is None:
            dx = np.ones_like(x)
        xspan = x[-1] - x[0]
        mu = x[0] + xspan * center_ratio
        msk = (x >= (mu - xspan * width_ratio)) & (x <= mu + xspan * width_ratio)

        # TODO: update this, minimizer creates NaN's if default sigma used (0)
        sigma = xspan * width_ratio / 10.0
        amp = np.max(y[msk]) * np.sqrt(2 * np.pi) * sigma
        # TODO: We miss gamma here
        return [
            (f"{self.prefix}amp", "value", amp),
            (f"{self.prefix}amp", "min", 0.0),
            (f"{self.prefix}mu", "value", mu),
            (f"{self.prefix}mu", "min", x[0]),
            (f"{self.prefix}mu", "max", x[-1]),
            (f"{self.prefix}sigma", "value", sigma),
            (f"{self.prefix}sigma", "min", 0.0),
            (f"{self.prefix}gamma", "value", 0.95),
            (f"{self.prefix}gamma", "min", 0.0),
        ]


MODEL_STR_TO_CLS = {
    "constant": ConstantModel,
    "line": LineModel,
    "gauss": GaussModel,
    "gausserf": GaussErfModel,
    "erf": ErfModel,
    "exp": ExpModel,
    "expgauss": ExpGaussModel,
}


# -----------------------------------------------------------------------------
# Fitters
# TODO: add docs
# TODO: add ability to override defaults
# TODO: add ability to initialize and fit with Fitter.__init__
# TODO: include x_edges?
# TODO: handle y normalization (i.e. cps vs cps/keV), needs x_edges
# TODO: use set_param_hint to set global model defaults
# -----------------------------------------------------------------------------


class Fitter:
    """Base class for more specialized fit objects.

    A note on interpreting fit results: ascribing meaning to histogram fit
    parameters is notoriously tricky, since the y-scale has units of counts
    per bin width, not just counts. The user may need to divide by the
    histogram bin width for area- or height-like parameters if the histogram is
    not already normalized by bin width. See, e.g., p. 171 of Bevington and
    Robinson, "Data reduction and error analysis for the physical sciences".
    """

    def __init__(self, model, x=None, y=None, y_unc=None, dx=None, roi=None):
        # Initialize
        self._model = None
        self._name = None
        self._x = None
        self._y = None
        self._y_unc = None
        self._roi = None
        self._roi_msk = None
        self._xmode = None
        self._ymode = None
        self.result = None
        self.dx = None
        # Model and parameters
        self._make_model(model)
        self.params = self.model.make_params()
        # Set data
        self.set_data(x=x, y=y, y_unc=y_unc, dx=dx, roi=roi, update_defaults=True)

    def __str__(self):
        return (
            "bq.Fitter instance\n"
            + f"     name: {self.name}\n"
            + f"    model: {self.model}\n"
            + f"        x: {self.x}\n"
            + f"        y: {self.y}\n"
            + f"    y_unc: {self.y_unc}\n"
            + f"    xmode: {self.xmode}\n"
            + f"    ymode: {self.ymode}\n"
            + f"       dx: {self.dx}\n"
            + f"      roi: {self.roi}"
        )

    __repr__ = __str__

    @property
    def name(self):
        return self._name

    @property
    def model(self):
        return self._model

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def y_unc(self):
        if self._y_unc is None:
            warnings.warn(
                "No y uncertainties (y_unc) provided. The fit will not be "
                + "weighted causing in poor results at low counting statistics.",
                FittingWarning,
            )
        return self._y_unc

    @y_unc.setter
    def y_unc(self, y_unc):
        if y_unc is not None:
            self._y_unc = np.asarray(y_unc, dtype=float)
            assert len(self.x) == len(
                self._y_unc
            ), "Fitting x (len {}) does not match y_unc (len {})".format(
                len(self.x), len(self._y_unc)
            )
            if np.any(self._y_unc <= 0.0):
                min_v = np.min(self._y_unc[self._y_unc > 0.0])
                warnings.warn(
                    "Negative or zero uncertainty not supported. Changing "
                    + f"them to {min_v}. If you have Poisson data, "
                    + "this should be 1."
                )
                self._y_unc[self._y_unc <= 0.0] = min_v
        else:
            self._y_unc = None

    @property
    def x_roi(self):
        return self.x[self.roi_msk]

    @property
    def y_roi(self):
        return self.y[self.roi_msk]

    @property
    def y_unc_roi(self):
        if self.y_unc is None:
            return None
        return self.y_unc[self.roi_msk]

    @property
    def dx_roi(self):
        if self.dx is None:
            return None
        return self.dx[self.roi_msk]

    @property
    def roi(self):
        return self._roi

    @property
    def roi_msk(self):
        if self._roi_msk is None:
            return np.ones_like(self.x, dtype=bool)
        else:
            return self._roi_msk

    @property
    def xmode(self):
        return self._xmode

    @property
    def ymode(self):
        return self._ymode

    @property
    def param_names(self):
        return list(self.params.keys())

    def set_data(self, y, x=None, y_unc=None, dx=None, roi=None, update_defaults=True):
        # Set y data (skip if y is None)
        if y is None:
            return
        self._y = np.asarray(y)
        # Set x data
        if x is None:
            self._x = np.arange(len(self.y))
        else:
            self._x = np.asarray(x)
            assert len(self.x) == len(
                self.y
            ), "Fitting x (len {}) does not match y (len {})".format(
                len(self.x), len(self.y)
            )
        # Handle y uncertainties
        self.y_unc = y_unc
        # set deltax (bin width)
        self.dx = dx

        if roi is not None:
            self.set_roi(*roi)
        if update_defaults:
            self.guess_param_defaults(update=True)

    def set_roi(self, low, high, update_defaults=True):
        """Set the region of interest (ROI) of x-values for the fit.

        Parameters
        ----------
        low : float
            Lower x-value of the ROI
        high : float
            Upper x-value of the ROI
        update_defaults : bool, optional
            If True, recompute default params based on new ROI.
        """

        self._roi = (float(low), float(high))
        self._roi_msk = (self.x >= self.roi[0]) & (self.x <= self.roi[1])
        if update_defaults:
            self.guess_param_defaults(update=True)

    def set_param(self, pname, ptype, pvalue):
        self.params[pname].set(**{ptype: pvalue})

    def _translate_model(self, m):
        if inspect.isclass(m):
            if not issubclass(m, Model):
                raise FittingError(f"Input model is not a subclass of Model: {m}")
            self._model_cls_cnt[m] = self._model_cls_cnt.get(m, 0) + 1
            return m
        elif isinstance(m, Model):
            cls = m.__class__
            self._model_cls_cnt[cls] = self._model_cls_cnt.get(cls, 0) + 1
            return m
        elif isinstance(m, str):
            cls = MODEL_STR_TO_CLS.get(m.lower(), None)
            if cls is not None:
                self._model_cls_cnt[cls] = self._model_cls_cnt.get(cls, 0) + 1
                return cls
        raise FittingError(f"Unknown model type: {m}")

    def _make_model(self, model):
        if isinstance(model, str) or isinstance(model, Model):
            model = [model]
        # Convert the model(s) to a list of Model classes / Model instancess
        self._model_cls_cnt = {}
        model_translated = [self._translate_model(m) for m in model]
        # Build complete model with appropriate prefixes
        model_prefixes = set()
        models = []
        name = ""
        for m in model_translated:
            if inspect.isclass(m):
                prefix_base = m.__name__.lower()
                if prefix_base.endswith("model"):
                    prefix_base = prefix_base[:-5]
                if self._model_cls_cnt[m] == 1:
                    prefix = f"{prefix_base}_"
                else:
                    for i in range(self._model_cls_cnt[m]):
                        prefix = f"{prefix_base}{i}_"
                        if prefix not in model_prefixes:
                            break
                m_instance = m(prefix=prefix)
            else:
                m_instance = m
            if m_instance.prefix in model_prefixes:
                raise FittingError(
                    "A model prefix is not unique: "
                    + f"{m_instance.prefix} "
                    + f"All models: {model_translated}"
                )
            model_prefixes.add(m_instance.prefix)
            models.append(m_instance)
            name += m_instance._name.capitalize()
        # Construct final model
        self._name = name
        self._model = models[0]
        for m in models[1:]:
            self._model += m

    def _guess_param_defaults(self, **kwargs):
        params = []
        for comp in self.model.components:
            params += comp.guess(self.y_roi, x=self.x_roi, dx=self.dx_roi)
        return params

    def guess_param_defaults(self, update=False, **kwargs):
        defaults = self._guess_param_defaults(**kwargs)
        if update:
            # TODO: check this logic
            if defaults is not None:
                for dp in defaults:
                    self.set_param(*dp)
        return defaults

    def fit(self, backend="lmfit", guess=None, limits=None):
        """Perform the weighted fit to data.

        Parameters
        ----------
        backend : {'lmfit', 'lmfit-pml', '[i]minuit-pml'}
            Backend fitting module to use.
        guess : dict of {str: numeric}, optional
            User-specified parameter guesses that override guesses provided by
            the Model.guess() or Fitter.guess_param_defaults() methods.
            Currently only implemented for backend="lmfit" and "iminuit".
        limits : dict of {str: tuple}, optional
            User-specified parameter limits that override limits provided by
            the Model.guess() or Fitter.guess_param_defaults() methods.
            Currently only implemented for backend="lmfit" and "iminuit".

        Raises
        ------
        FittingError
            If `backend` is not supported.
        AssertionError
            If self.y is None.
        """

        assert self.y is not None, "No data initialized, did you call set_data?"
        self.result = None
        y_roi_norm = self.y_roi
        if self.dx is not None:
            y_roi_norm = self.y_roi / self.dx_roi

        self.backend = backend.lower().strip()

        if self.backend == "lmfit":
            # Perform the fit, weighted by 1/uncertainties.
            weights = self.y_unc_roi ** -1.0
            self.result = self.model.fit(
                y_roi_norm, self.params, x=self.x_roi, weights=weights
            )

        elif self.backend == "lmfit-pml":
            self._set_likelihood_residual()
            # Perform the fit. PML automatically applies 1/sqrt(y) weights, so
            # additional weights here just convert back to counts.
            self.result = self.model.fit(
                self.y_roi,
                self.params,  # self.result.params,
                x=self.x_roi,
                weights=self.dx_roi,
                fit_kws={"reduce_fcn": lambda r: np.sum(r)},
                method="Nelder-Mead",
                calc_covar=False,
            )  # no, bounds, default would be L-BFGS-B'
            # NOTE: Calculating errors in lmfit-pml breaks minimization

        elif self.backend in ["iminuit", "minuit"]:
            raise NotImplementedError(
                f"Backend {self.backend} with least-squares loss not yet "
                + f"supported. Use {self.backend}-pml for Poisson loss or "
                + "lmfit for least-squares."
            )

        elif self.backend in ["iminuit-pml", "minuit-pml"]:
            # Translate a model from lmfit to minuit

            if not _is_count_like(self.y_roi):
                warnings.warn(
                    "Passing non-count-like data to a Poisson loss fit", FittingWarning
                )

            def model_loss(*args):
                """Poisson loss given the model specified by args."""
                # Convert args to kwargs as lmfit.model.eval _requires_ kwargs,
                # while the cost function passed to the Minuit object seems to
                # _require_ args.
                kwargs = {self.model.param_names[i]: arg for i, arg in enumerate(args)}
                y_eval = self.model.eval(x=self.x_roi, **kwargs)
                if self.dx_roi is not None:
                    y_eval *= self.dx_roi
                return poisson_loss(y_eval, self.y_roi)

            # Filter out fixed params and have one consistent variable name
            # instead of all the params / param_names / parameters / etc.
            free_vars = [p for p in self.params if self.params[p].vary]

            # Parameter guesses
            try:
                # First try parsing the lmfit guess into iminuit
                guess_l = self.guess_param_defaults()
                guess_i = {
                    g[0]: g[2]
                    for g in guess_l
                    if (g[1] == "value" and g[0] in free_vars)
                }
            except NotImplementedError:
                # If the model/component does not have a guess() method
                guess_i = {}
            # Then override with user input
            if guess is not None:
                guess_i.update(guess)
                del guess  # eliminate subscript confusion

            # Handle user parameter limits
            # Supplying guesses and/or bounds seems more important for minuit
            try:
                # First try parsing the lmfit limits into iminuit
                limits_l = self.guess_param_defaults()
                min_vals, max_vals = {}, {}
                for lim in limits_l:
                    if lim[1] == "min":
                        min_vals[lim[0]] = lim[2]
                    elif lim[1] == "max":
                        max_vals[lim[0]] = lim[2]
                limits_i = {
                    p: (min_vals.get(p, None), max_vals.get(p, None)) for p in free_vars
                }
            except NotImplementedError:
                # If the model/component does not have a guess() method
                limits_i = {}
            # Then override with user input
            if limits is not None:
                limits_i.update(limits)
                del limits  # eliminate subscript confusion

            # Since Minuit requires guesses for every parameter, if we don't
            # have a guess, use limits_i midpoint, or zero barring that.
            for p in free_vars:
                if p not in guess_i:
                    warn_str = f"No guess provided for parameter {p}. "
                    if (
                        p in limits
                        and limits_i[p][0] is not None
                        and limits_i[p][1] is not None
                    ):
                        warnings.warn(warn_str + "Setting to limits midpoint.")
                        guess_i[p] = 0.5 * (limits_i[p][0] + limits_i[p][1])
                    else:
                        warnings.warn(warn_str + "Setting to 0.")
                        guess_i[p] = 0.0

            # Set up the Minuit minimizer with initial guess
            self.result = Minuit(model_loss, name=free_vars, **guess_i)

            # Set user parameter limits
            for k, v in limits_i.items():
                self.result.limits[k] = v

            # Specify proper error definition for likelihood model
            self.result.errordef = Minuit.LIKELIHOOD

            # Specify extra checks
            self.result.strategy = 2

            # Run the minimizer
            self.result.migrad()

            # Compute errors
            self.result.hesse()
            try:
                self.result.minos()
            except RuntimeError as exc:
                warnings.warn(f"Could not compute Minos errors: {exc}")

            # Update the interfaces to best_values and init_values
            self._best_values, self._init_values = {}, {}
            for i in range(self.result.npar):
                p = self.result.parameters[i]
                self._best_values[p] = self.result.values[i]
                self._init_values[p] = self.result.init_params[p].value

            # Arg order sanity checks
            assert list(self._best_values.keys()) == free_vars
            assert list(self._init_values.keys()) == free_vars
            assert list(self.result.parameters) == free_vars

        else:
            raise FittingError(f"Unknown fitting backend: {backend}")

    def _set_likelihood_residual(self):
        def _likelihood_residual(self, params, data, weights, **kwargs):
            """same as model._residual of lmfit"""
            model = self.eval(params, **kwargs)
            if weights is not None:
                model *= weights
            if self.nan_policy == "raise" and not np.all(np.isfinite(model)):
                msg = (
                    "The model function generated NaN values and the fit "
                    "aborted! Please check your model function and/or set "
                    "boundaries on parameters where applicable. In cases "
                    "like this, using \"nan_policy='omit'\" will probably "
                    "not work."
                )
                raise ValueError(msg)
            mask = model <= 0  # This should not be necessary
            diff = model - scipy.special.xlogy(data, model)
            diff[mask] = 1e32
            if diff.dtype == np.complex:
                # data/model are complex
                diff = diff.ravel().view(float)
            return np.asarray(diff).ravel()  # for compatibility with pandas.Series

        # This overwrites the  model residual method, is an ugly hack to make
        # poisson fitting possible. This is not undone for now.
        self.model._residual = _likelihood_residual.__get__(self.model, Model)

    def eval(self, x, params=None, **kwargs):
        return self.model.eval(x=x, params=params, **kwargs)

    def param_val(self, param):
        """
        Value of fit parameter `param`
        """
        if self.result is None:
            return None
        if "lmfit" in self.backend:
            if param in self.result.params:
                return self.result.params[param].value
            elif param in self.fit.best_values:
                return self.result.best_values[param]
            else:
                raise FittingError("Unknown param: {}", param)
        elif "minuit" in self.backend:
            return self.result.params[param].value
        else:
            raise FittingError("Unknown backend: {}", self.backend)

    def param_unc(self, param):
        """
        Fit error of fit parameter `param`
        """
        if self.result is None:
            return None
        if "lmfit" in self.backend:
            if param in self.result.params:
                return self.result.params[param].stderr
            elif param in self.result.best_values:
                # This is the case for the `erf_form` key
                return np.nan
            else:
                raise FittingError("Unknown param: {}", param)
        elif "minuit" in self.backend:
            return self.result.params[param].error  # TODO minos vs hesse?
        else:
            raise FittingError("Unknown backend: {}", self.backend)

    @property
    def best_values(self):
        """Wrapper for dictionary of best_values."""
        if "lmfit" in self.backend:
            return self.result.best_values
        elif "minuit" in self.backend:
            return self._best_values
        else:
            raise FittingError("Unknown backend: {}", self.backend)

    @property
    def init_values(self):
        """Wrapper for dictionary of init_values."""
        if "lmfit" in self.backend:
            return self.result.init_values
        elif "minuit" in self.backend:
            return self._init_values
        else:
            raise FittingError("Unknown backend: {}", self.backend)

    @property
    def success(self):
        if "lmfit" in self.backend:
            return self.result.success
        elif "minuit" in self.backend:
            return self.result.valid
        else:
            raise FittingError("Unknown backend: {}", self.backend)

    def param_dataframe(self, sort_by_model=False):
        """
        Dataframe of all fit parameters value and fit error
        """
        if self.result is None:
            return None
        df = pd.DataFrame(columns=["val", "unc"], dtype=float)
        for k in self.param_names:
            df.loc[k, "val"] = self.param_val(k)
            df.loc[k, "unc"] = self.param_unc(k)
        if sort_by_model:
            df.set_index(
                pd.MultiIndex.from_tuples(
                    [tuple(p.split("_")) for p in df.index], names=["model", "param"]
                ),
                inplace=True,
            )
        return df

    def compute_residuals(self, residual_type="abs"):
        """Compute residuals between the data and the fit.

        Parameters
        ----------
        residual_type : {'abs', 'rel', 'sigma'}, optional
            Residual type to calculate (default: 'abs')
                'abs' : data - fit
                'rel' : (data - fit) / |fit|
                'sigma' : (data - fit) / (data_uncertainty)

        Returns
        -------
        np.ndarray
            Array of residuals
        """
        dx_roi = np.ones_like(self.x_roi) if self.dx_roi is None else self.dx_roi
        y_eval = self.eval(self.x_roi, **self.best_values) * dx_roi
        y_res = self.y_roi - y_eval

        if residual_type == "rel":
            # Residuals relative to the model evaluation
            return y_res / np.abs(y_eval)
        elif residual_type == "sigma":
            # Residuals relative to the data uncertainty
            return y_res / self.y_unc_roi
        elif residual_type == "abs":
            # Absolute residuals
            return y_res
        else:
            raise ValueError(f"Unknown residuals type: {residual_type:s}")

    def plot(self, npts=1000, **kwargs):
        """Plot the fit result on the current axis.

        Parameters
        ----------
        npts : int (optional)
            Number of points in x to generate.
        kwargs : dict (optional)
            Additional kwargs passed to plt.plot().

        Returns
        -------
        int
            Description of anonymous integer return value.
        """

        x_plot = np.linspace(self.x_roi[0], self.x_roi[-1], npts)
        y = self.eval(x_plot, **self.best_values)
        plt.plot(x_plot, y, **kwargs)

    def custom_plot(
        self,
        title=None,
        savefname=None,
        title_fontsize=24,
        title_fontweight="bold",
        residual_type="abs",
        **kwargs,
    ):
        """Three-panel figure showing fit results.

        Top-left panel shows the data and the fit. Bottom-left shows the fit
        residuals. Right prints fit statistics and correlations.

        Parameters
        ----------
        title : str, optional
            Title of the figure (default: no title)
        savefname : str, optional
            Filename to save the figure as (default: not saved)
        title_fontsize : int, optional
            Title font size (default: 24)
        title_fontweight : str, optional
            Title font weight (default: 'bold')
        residual_type : {'abs', 'rel', 'sigma'}, optional
            Residual type to calculate (default: 'abs')
                'abs' : data - fit
                'rel' : (data - fit) / |fit|
                'sigma' : (data - fit) / (data_uncertainty)
        **kwargs
            Additional kwargs. Currently unused.

        Returns
        -------
        matplotlib figure
            Returned only if savefname is None
        """

        ymin, ymax = self.y_roi.min(), self.y_roi.max()
        # Prepare plots
        dx, dx_roi = self.dx, self.dx_roi
        if dx is None:
            dx = np.ones_like(self.x)
        if dx_roi is None:
            dx_roi = np.ones_like(self.x_roi)
        gs = GridSpec(2, 2, height_ratios=(4, 1))
        gs.update(
            left=0.05, right=0.99, wspace=0.03, top=0.94, bottom=0.06, hspace=0.06
        )
        fig = plt.figure(figsize=(18, 9))
        fit_ax = fig.add_subplot(gs[0, 0])
        res_ax = fig.add_subplot(gs[1, 0], sharex=fit_ax)
        txt_ax = fig.add_subplot(gs[:, 1])
        txt_ax.get_xaxis().set_visible(False)
        txt_ax.get_yaxis().set_visible(False)
        # Set fig title
        if title is not None:
            fig.suptitle(
                str(title), fontweight=title_fontweight, fontsize=title_fontsize
            )

        # ---------------------------------------
        # Fit plot (keep track of min/max in roi)
        # ---------------------------------------
        # Smooth roi x values
        x_plot = np.linspace(self.x_roi[0], self.x_roi[-1], 1000)
        # All data (not only roi)
        fit_ax.errorbar(
            self.x,
            self.y / dx,
            yerr=self.y_unc,
            c="k",
            fmt="o",
            markersize=5,
            alpha=0.1,
            label="data",
        )
        # Init fit
        y = self.eval(x_plot, **self.init_values)
        ymin, ymax = min(y.min(), ymin), max(y.max(), ymax)
        fit_ax.plot(x_plot, y, "k--", label="init")
        # Best fit
        y = self.eval(x_plot, **self.best_values)
        ymin, ymax = min(y.min(), ymin), max(y.max(), ymax)
        fit_ax.plot(x_plot, y, color="#e31a1c", label="best fit", zorder=10)
        # Plot 1 sigma uncertainty bands
        # TODO: for minuit, see how_to_draw_error_bands iminuit notebook
        if self.success and "lmfit" in self.backend:
            yunc = self.result.eval_uncertainty(x=x_plot, sigma=1)
            fit_ax.fill_between(
                x_plot,
                y - yunc,
                y + yunc,
                label=r"$\pm 1\, \sigma$",
                color="#e31a1c",
                alpha=0.2,
                zorder=9,
            )
        # Components
        fit_ax.set_prop_cycle(color=COLORS)
        for i, m in enumerate(self.model.components):
            y = m.eval(x=x_plot, **self.best_values)
            if isinstance(y, float):
                y = np.ones(x_plot.shape) * y
            ymin, ymax = min(y.min(), ymin), max(y.max(), ymax)
            fit_ax.plot(x_plot, y, label=m.prefix)
        # Plot Peak center and FWHM
        peak_centers = [
            self.param_val(p)
            for p in self.param_names
            if (p.startswith("gauss") and p.endswith("mu"))
        ]
        peak_fwhms = [
            self.param_val(p) * FWHM_SIG_RATIO
            for p in self.param_names
            if (p.startswith("gauss") and p.endswith("sigma"))
        ]
        for i, (c, f) in enumerate(zip(peak_centers, peak_fwhms)):
            if i == 0:
                label = "Centroid and FWHM"
            else:
                label = None
            fit_ax.axvline(c, color="#ff7f00")
            fit_ax.axvspan(
                c - f / 2.0, c + f / 2.0, color="#ff7f00", alpha=0.2, label=label
            )
        # Misc
        fit_ax.legend(loc="upper right")
        # Set viewing window to only include the roi (not entire spectrum)
        xpad = (self.x_roi[-1] - self.x_roi[0]) * 0.05
        ypad = (ymax - ymin) * 0.05
        fit_ax.set_xlim([self.x_roi[0] - xpad, self.x_roi[-1] + xpad])
        fit_ax.set_ylim([ymin - ypad, ymax + ypad])
        fit_ax.set_ylabel(self.ymode)

        # ---------
        # Residuals
        # ---------
        y_eval = self.eval(self.x_roi, **self.best_values) * dx_roi
        res_kwargs = dict(fmt="o", color="k", markersize=5, label="residuals")

        # Y-values of the residual plot, depending on residual_type
        y_plot = self.compute_residuals(residual_type)

        # Error bars and ylabel of the residual plot
        if residual_type == "rel":
            yerr_plot = self.y_unc_roi / np.abs(y_eval)
            ylabel = "Relative residuals"
        elif residual_type == "sigma":
            yerr_plot = np.zeros_like(y_plot)
            ylabel = r"Residuals $(\sigma)$"
        elif residual_type == "abs":
            yerr_plot = self.y_unc_roi
            ylabel = "Residuals"
        else:
            raise ValueError(f"Unknown residuals option: {residual_type:s}")
        res_ax.errorbar(x=self.x_roi, y=y_plot, yerr=yerr_plot, **res_kwargs)
        res_ax.axhline(0.0, linestyle="dashed", c="k", linewidth=1.0)
        res_ax.set_ylabel(ylabel)
        res_ax.set_xlabel(self.xmode)

        # -------------------
        # Fit report (txt_ax)
        # -------------------
        props = dict(boxstyle="round", facecolor="white", edgecolor="black", alpha=1)
        props = dict(facecolor="white", edgecolor="none", alpha=0)
        fp = FontProperties(family="monospace", size=8)
        if "lmfit" in self.backend:
            best_fit_values = ""
            op = self.result.params
            for p in self.result.params:
                if op[p].stderr is None:
                    pass
                    # TODO: Calculate errors breaks minimization right now
                    # warnings.warn(
                    #     "Package numdifftools is required to have "
                    #     "stderr calculated.", FittingWarning)
                else:
                    best_fit_values += "{:15} {: .6e} +/- {:.5e} ({:6.1%})\n".format(
                        p, op[p].value, op[p].stderr, abs(op[p].stderr / op[p].value)
                    )
            best_fit_values += "{:15} {: .6e}\n".format(
                "Chi Squared:", self.result.chisqr
            )
            best_fit_values += "{:15} {: .6e}".format(
                "Reduced Chi Sq:", self.result.redchi
            )
            # Remove first 2 lines of fit report (long model description)
            s = "\n".join(self.result.fit_report().split("\n")[2:])
            # Add some more parameter details
            s += "\n"
            param_df = self.param_dataframe(sort_by_model=True)
            for model_name, sdf in param_df.groupby(level="model"):
                s += model_name + "\n"
                for (_, param_name), param_data in sdf.iterrows():
                    v = param_data["val"]
                    e = param_data["unc"]
                    s += "    {:24}: {: .6e} +/- {:.5e} ({:6.1%})\n".format(
                        param_name, v, e, np.abs(e / v)
                    )
        elif "minuit" in self.backend:
            s = str(self.result) + "\n"

        # Add info about the ROI and units
        if self.roi:
            s += "ROI: [{:.3f}, {:.3f}]\n".format(*self.roi)
        s += "X units: {:s}\n".format(self.xmode if self.xmode else "None")
        s += "Y units: {:s}\n".format(self.ymode if self.ymode else "None")
        # Add to empty axis
        txt_ax.text(
            x=0.01,
            y=0.99,
            s=s,
            fontproperties=fp,
            ha="left",
            va="top",
            transform=txt_ax.transAxes,
            bbox=props,
        )

        if savefname is not None:
            fig.savefig(savefname)
            plt.close(fig)
        else:
            return fig



#!/usr/bin/env python
"""Becquerel: Tools for radiation spectral analysis."""

import sys
import site
from setuptools import setup, find_packages
import importlib.util

# Enables --editable install with --user
# https://github.com/pypa/pip/issues/7953
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

_spec = importlib.util.spec_from_file_location(
    "__metadata__", "./becquerel/__metadata__.py"
)
METADATA = importlib.util.module_from_spec(_spec)
_spec.loader.exec_module(METADATA)

# Enables --editable install with --user
# https://github.com/pypa/pip/issues/7953
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

# remove package title from description
with open("README.md") as fh:
    README = "\n".join(fh.readlines()[2:])

with open("CONTRIBUTING.md") as fh:
    CONTRIBUTING = fh.read()

with open("requirements.txt") as fh:
    REQUIREMENTS = [_line for _line in fh.readlines() if _line]

with open("requirements-dev.txt") as fh:
    REQUIREMENTS_DEV = [l.strip() for l in fh.readlines() if not l.startswith("-r")]

# make long description from README and CONTRIBUTING
# but move copyright notice to the end
LONG_DESCRIPTION = "{0}\n{2}\n## Copyright Notice\n{1}".format(
    *README.split("## Copyright Notice"), CONTRIBUTING
)

setup(
    name=METADATA.__name__,
    version=METADATA.__version__,
    description=METADATA.__description__,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=METADATA.__url__,
    download_url=f"{METADATA.__url__}/releases",
    maintainer=METADATA.__maintainer__,
    maintainer_email=METADATA.__email__,
    classifiers=METADATA.__classifiers__,
    platforms="any",
    packages=find_packages(),
    python_requires=">=3.6",
    install_requires=REQUIREMENTS,
    setup_requires=["pytest-runner"],
    license=METADATA.__license__,
    tests_require=REQUIREMENTS_DEV,
)



"""Query half-life and decay data from the National Nuclear Data Center.

References:
  http://www.nndc.bnl.gov
  http://www.nndc.bnl.gov/nudat2/indx_sigma.jsp
  http://www.nndc.bnl.gov/nudat2/indx_dec.jsp

"""

from future.builtins import super
import warnings
import numpy as np
import requests
import pandas as pd
import uncertainties


PARITIES = ["+", "-", "any"]


DECAYRAD_DECAY_MODE = {
    "any": "ANY",
    "internal transition": "IT",
    "it": "IT",
    "beta-": "B-",
    "b-": "B-",
    "electron capture beta+": "ECBP",
    "ecbp": "ECBP",
    "ecb+": "ECBP",
    "ec+b+": "ECBP",
    "electron capture": "ECBP",
    "ec": "ECBP",
    "beta+": "ECBP",
    "b+": "ECBP",
    "neutron": "N",
    "n": "N",
    "proton": "P",
    "p": "P",
    "alpha": "A",
    "a": "A",
    "spontaneous fission": "SF",
    "sf": "SF",
}


WALLET_DECAY_MODE = dict(DECAYRAD_DECAY_MODE)
WALLET_DECAY_MODE.update(
    {
        "double beta": "DB",
        "bb": "DB",
        "cluster": "C",
        "c": "C",
        "beta-delayed neutron": "DN",
        "b-delayed n": "DN",
        "bdn": "DN",
        "beta-delayed proton": "DP",
        "b-delayed p": "DP",
        "bdp": "DP",
        "beta-delayed alpha": "DA",
        "b-delayed a": "DA",
        "bda": "DA",
        "beta-delayed fission": "DF",
        "b-delayed f": "DF",
        "bdf": "DF",
    }
)


DECAYRAD_RADIATION_TYPE = {
    "any": "ANY",
    "gamma": "G",
    "g": "G",
    "beta-": "BM",
    "b-": "BM",
    "beta+": "BP",
    "b+": "BP",
    "electron": "E",
    "e": "E",
    "proton": "P",
    "p": "P",
    "alpha": "A",
    "a": "A",
}


class NNDCError(Exception):
    """General NNDC request error."""

    pass


class NoDataFound(NNDCError):
    """No datasets were found within the specified search."""


class NNDCInputError(NNDCError):
    """Error related to the user input to an NNDC query."""

    pass


class NNDCRequestError(NNDCError):
    """Error related to communicating with NNDC or parsing the result."""

    pass


def _parse_headers(headers):
    """Parse table headers and ensure they are unique.

    Args:
      headers: a list of column header strings.

    Returns:
      a new list of strings where abbreviations have been expanded.

    Raises:
      NNDCRequestError: if there was a problem parsing the headers.

    """

    headers_new = []
    # reformat column headers if needed
    for j, hd in enumerate(headers):
        # rename so always have T1/2 (s)
        if hd == "T1/2 (num)" or hd == "T1/2 (seconds)":
            hd = "T1/2 (s)"
        # for uncertainties, add previous column header to it
        if j > 0 and "Unc" in hd:
            hd = headers[j - 1] + " " + hd
        if "Unc" in hd and "Unc." not in hd:
            hd = hd.replace("Unc", "Unc.")
        # expand abbreviated headers
        if "Energy" in hd and "Energy Level" not in hd:
            hd = hd.replace("Energy", "Energy Level")
        if "Par. Elevel" in hd:
            hd = hd.replace("Par. Elevel", "Parent Energy Level")
        if "Abund." in hd:
            hd = hd.replace("Abund.", "Abundance (%)")
        if "Ene." in hd:
            hd = hd.replace("Ene.", "Energy")
        if "Int." in hd:
            hd = hd.replace("Int.", "Intensity (%)")
        if "Dec" in hd and "Decay" not in hd:
            hd = hd.replace("Dec", "Decay")
        if "Rad" in hd and "Radiation" not in hd:
            hd = hd.replace("Rad", "Radiation")
        if "EP" in hd:
            hd = hd.replace("EP", "Endpoint")
        if "Mass Exc" in hd and "Mass Excess" not in hd:
            hd = hd.replace("Mass Exc", "Mass Excess")
        headers_new.append(hd)
    if len(set(headers_new)) != len(headers_new):
        raise NNDCRequestError(
            "Duplicate headers after parsing\n"
            + f'    Original headers: "{headers}"\n'
            + f'    Parsed headers:   "{headers_new}"'
        )
    return headers_new


def _parse_table(text):
    """Parse table contained in the text into a dictionary.

    Args:
      text: a string containing an HTML table from the NNDC request

    Returns:
      a dictionary of lists keyed by the column headers.

    Raises:
      NNDCRequestError: if unable to parse the table.

    """

    text = str(text)
    try:
        text = text.split("<pre>")[1]
        text = text.split("</pre>")[0]
        text = text.split("To save this output")[0]
        lines = text.split("\n")
    except Exception as exc:
        raise NNDCRequestError(f"Unable to parse text:\n{exc}\n{text}")
    table = {}
    headers = None
    for line in lines:
        tokens = line.split("\t")
        tokens = [t.strip() for t in tokens]
        if len(tokens) <= 1:
            continue
        if headers is None:
            headers = tokens
            headers = _parse_headers(headers)
            for header in headers:
                table[header] = []
        else:
            if len(tokens) != len(headers):
                raise NNDCRequestError(
                    "Too few data in table row\n"
                    + f'    Headers: "{headers}"\n'
                    + f'    Row:     "{tokens}"'
                )
            for header, token in zip(headers, tokens):
                table[header].append(token)
    return table


def _parse_float_uncertainty(x, dx):
    """Parse a string and its uncertainty into a float or ufloat.

    Examples:
      >>> _parse_float_uncertainty('257.123', '0.005')
      257.123+/-0.005
      >>> _parse_float_uncertainty('8', '')
      8.0

    Args:
      x: a string representing the nominal value of the quantity.
      dx: a string representing the uncertainty of the quantity.

    Returns:
      a float (if dx == '') or a ufloat.

    Raises:
      NNDCRequestError: if values cannot be parsed.

    """

    if not isinstance(x, str):
        raise NNDCRequestError(f"Value must be a string: {x}")
    if not isinstance(dx, str):
        raise NNDCRequestError(f"Uncertainty must be a string: {dx}")
    # ignore percents
    if "%" in x:
        x = x.replace("%", "")
    # ignore unknown ground state levels (X, Y, Z, W)
    for sym in ["X", "Y", "Z", "W"]:
        if "+" + sym in x:
            x = x.replace("+" + sym, "")
        elif x == sym:
            x = "0"
    # handle special ENSDF abbreviations, e.g.,
    # http://www.iaea.org/inis/collection/NCLCollectionStore/_Public/14/785/14785563.pdf
    # "One of the following expressions:
    #   LT, GT, LE, GE, AP, CA, SY
    # for less than, greater than, less than or equal to greater
    # than or equal to. approximately equal to, calculated, and
    # from systematics, respectively."
    for sym in ["*", "<", ">", "=", "~", "?", "@", "&", "P", "N"]:
        while sym in x:
            x = x.replace(sym, "")
    # correct specific typos in the database
    if "E-11 0" in x:
        x = x.replace("E-11 0", "E-11")
    if "E-12 0" in x:
        x = x.replace("E-12 0", "E-12")
    if "0.0000 1" in x:
        x = x.replace("0.0000 1", "0.0000")
    if "2 .8E-7" in x:
        x = x.replace("2 .8E-7", "2.8E-7")
    if "8 .0E-E5" in x:
        x = x.replace("8 .0E-E5", "8.0E-5")
    # handle blank or missing data
    if x == "" or x == " ":
        return None
    if "****" in dx:
        dx = ""
    elif dx in ["LT", "GT", "LE", "GE", "AP", "CA", "SY"]:
        dx = ""
    try:
        x2 = float(x)
    except ValueError:
        raise NNDCRequestError(f'Value cannot be parsed as float: "{x}"')
    if dx == "":
        return x2
    # handle multiple exponents with some uncertainties, e.g., "7E-4E-5"
    tokens = dx.split("E")
    if len(tokens) == 3:
        dx = "E".join(tokens[:2])
        factor = pow(10.0, int(tokens[2]))
    else:
        factor = 1.0
    try:
        dx2 = float(dx) * factor
    except ValueError:
        raise NNDCRequestError(f'Uncertainty cannot be parsed as float: "{dx}"')
    return uncertainties.ufloat(x2, dx2)


def _format_range(x_range):
    """Return two strings for the two range elements, blank if not finite.

    Args:
      x_range: an iterable of 2 range limits, which can be numbers
        or inf/NaN/None.

    Returns:
      an iterable of 2 strings.

    Raises:
      NNDCInputError: if x_range is not an iterable of length 2.

    """

    try:
        x1, x2 = x_range
    except (TypeError, ValueError):
        raise NNDCInputError(f'Range keyword arg must have two elements: "{x_range}"')
    try:
        if np.isfinite(x1):
            x1 = f"{x1}"
        else:
            x1 = ""
    except TypeError:
        x1 = ""
    try:
        if np.isfinite(x2):
            x2 = f"{x2}"
        else:
            x2 = ""
    except TypeError:
        x2 = ""
    return x1, x2


class _NNDCQuery:
    """National Nuclear Data Center database query base class.

    Args:
      perform: a boolean dictating whether to immediately perform the query.
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    _URL = ""
    _DATA = {
        "spnuc": "",  # specify parent ('name', 'zan', or 'zanrange')
        "nuc": "",  # isotope name (use with 'name')
        "z": "",  # Z or element (use with 'zan')
        "zmin": "",  # Z min        (use with 'zanrange')
        "zmax": "",  # Z max        (use with 'zanrange')
        "a": "",  # A            (use with 'zan')
        "amin": "",  # A min        (use with 'zanrange')
        "amax": "",  # A max        (use with 'zanrange')
        "n": "",  # N            (use with 'zan')
        "nmin": "",  # N min        (use with 'zanrange')
        "nmax": "",  # N max        (use with 'zanrange')
        "evenz": "",  # 'any', 'even', or 'odd' Z (use with zanrange)
        "evena": "",  # 'any', 'even', or 'odd' A (use with zanrange')
        "evenn": "",  # 'any', 'even', or 'odd' N (use with zanrange)
        "tled": "disabled",  # half-life condition on/off
        "tlmin": "0",  # half-life min
        "utlow": "S",  # half-life min units ('S' = seconds)
        "tlmax": "3E17",  # half-life max
        "utupp": "S",  # half-life max units ('ST' = stable, 'GY' = Gy)
        "notlim": "disabled",  # half-life: no limit
        "dmed": "disabled",  # decay mode condition on/off
        "dmn": "ANY",  # decay mode: 'ANY' = any
        "out": "file",  # output to formatted file
        "unc": "stdandard",  # standard style uncertainties
        "sub": "Search",  # search for the data
    }
    _ALLOWED_KEYWORDS = [
        "perform",
        "nuc",
        "z",
        "a",
        "n",
        "z_range",
        "a_range",
        "n_range",
        "z_any",
        "z_even",
        "z_odd",
        "a_any",
        "a_even",
        "a_odd",
        "n_any",
        "n_even",
        "n_odd",
        "t_range",
    ]
    _DUMMY_TEXT = ""

    def __init__(self, **kwargs):
        """Initialize query of NNDC data."""
        perform = kwargs.get("perform", True)
        kwargs["perform"] = False
        self._data = dict(self._DATA)
        self._text = self._DUMMY_TEXT
        self.df = pd.DataFrame()
        self.update(**kwargs)
        if perform:
            self.perform()

    def __len__(self):
        """Length of any one of the data lists."""
        if self.df is None:
            return 0
        elif len(self.df.keys()) == 0:
            return 0
        else:
            return len(self.df[self.df.keys()[0]])

    def keys(self):
        """Return the data keys."""
        return self.df.keys()

    def __getitem__(self, key):
        """Return the list given by the key."""
        return self.df[key]

    def __setitem__(self, key, value):
        """Set the list given by the key."""
        self.df[key] = value

    def __str__(self):
        """Use str method for DataFrame."""
        return str(self.df)

    def __format__(self, formatstr):
        """Use format method for DataFrame."""
        return self.df.__format__(formatstr)

    def _request(self):
        """Request data table from the URL."""
        # Use the context manager to automatically cleanup the session
        with requests.Session() as session:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=ResourceWarning)
                resp = session.post(self._URL, data=self._data, stream=False)
            if not resp.ok or resp.reason != "OK" or resp.status_code != 200:
                raise NNDCRequestError("Request failed: " + resp.reason)
            for msg in [
                "Your search was unsuccessful",
                "Your search exceeded the maximum number of results",
                "There are too many results for your search",
            ]:
                if msg in resp.text:
                    raise NNDCRequestError("Request failed: " + msg)
            msg = "No datasets were found within the specified search"
            if msg in resp.text:
                raise NoDataFound(msg)
        return resp.text

    def update(self, **kwargs):
        """Update the search criteria."""
        for kwarg in kwargs:
            if kwarg not in self._ALLOWED_KEYWORDS:
                raise NNDCInputError(f'Unknown keyword: "{kwarg}"')
        if "nuc" in kwargs:
            self._data["spnuc"] = "name"
            self._data["nuc"] = kwargs["nuc"]
        for x in ["z", "a", "n"]:
            # handle Z, A, and N settings
            if x in kwargs:
                self._data["spnuc"] = "zanrange"
                self._data[x + "min"], self._data[x + "max"] = _format_range(
                    (kwargs[x], kwargs[x])
                )
            # handle *_range, *_any, *_odd, *_even
            elif x + "_range" in kwargs:
                self._data["spnuc"] = "zanrange"
                self._data[x + "min"], self._data[x + "max"] = _format_range(
                    kwargs[x + "_range"]
                )
                if self._data[x + "min"] == "":
                    self._data[x + "min"] = "0"
                if self._data[x + "max"] == "":
                    self._data[x + "max"] = "300"
            if x + "_any" in kwargs:
                self._data["even" + x] = "any"
            elif x + "_even" in kwargs:
                self._data["even" + x] = "even"
            elif x + "_odd" in kwargs:
                self._data["even" + x] = "odd"
        # handle half-life range condition
        if "t_range" in kwargs:
            self._data["tled"] = "enabled"
            self._data["tlmin"], self._data["tlmax"] = _format_range(kwargs["t_range"])

    def perform(self):
        """Perform the query."""
        # check the conditions
        if self._data["spnuc"] == "":
            self.update(z_range=(None, None))
        # submit the query
        try:
            self._text = self._request()
        except NoDataFound:
            self._text = self._DUMMY_TEXT
        if len(self._text) == 0:
            raise NNDCRequestError("NNDC returned no text")
        # package the output into a dictionary of arrays
        data = _parse_table(self._text)
        # create the DataFrame
        self.df = pd.DataFrame(data)
        # convert dimensionless integers to ints
        for col in ["A", "Z", "N", "M"]:
            if col in self.keys():
                self._convert_column(col, int)
        # combine uncertainty columns and add unit labels
        self._add_units_uncertainties()
        # add some more columns
        self._add_columns_energy_levels()
        # sort columns
        self._sort_columns()

    def _add_columns_energy_levels(self):
        """Add nuclear energy level 'M' and 'm' columns using energy levels."""
        if "Energy Level (MeV)" not in self.df:
            return
        # add column of integer M giving the isomer level (0, 1, 2, ...)
        self.df["M"] = [0] * len(self)
        # add string m giving the isomer level name (e.g., '' or 'm' or 'm2')
        self.df["m"] = [""] * len(self)
        # loop over each isotope in the dataframe
        A_Z = [(a, z) for a, z in zip(self["A"], self["Z"])]
        A_Z = set(A_Z)
        for a, z in A_Z:
            isotope = (self["A"] == a) & (self["Z"] == z)
            e_levels = []
            e_levels_nominal = []
            for e_level in self["Energy Level (MeV)"][isotope]:
                if isinstance(e_level, uncertainties.core.Variable):
                    e_level_nominal = e_level.nominal_value
                else:
                    e_level_nominal = e_level
                if e_level_nominal not in e_levels_nominal:
                    e_levels.append(e_level)
                    e_levels_nominal.append(e_level_nominal)
            e_levels = sorted(e_levels)
            for M, e_level in enumerate(e_levels):
                isomer = isotope & (abs(self["Energy Level (MeV)"] - e_level) < 1e-10)
                self.df.loc[isomer, "M"] = M
                if M > 0:
                    if len(e_levels) > 2:
                        self.df.loc[isomer, "m"] = f"m{M}"
                    else:
                        self.df.loc[isomer, "m"] = "m"

    def _add_units_uncertainties(self):
        """Add units and uncertainties with some columns as applicable."""
        if "Energy Level" in self.keys():
            self._convert_column(
                "Energy Level", lambda x: _parse_float_uncertainty(x, "")
            )
            self.df.rename(columns={"Energy Level": "Energy Level (MeV)"}, inplace=True)

        if "Parent Energy Level" in self.keys():
            self._convert_column_uncertainty("Parent Energy Level")
            self.df.rename(
                columns={"Parent Energy Level": "Energy Level (MeV)"}, inplace=True
            )
            self.df["Energy Level (MeV)"] *= 0.001

        if "Mass Excess" in self.keys():
            self._convert_column_uncertainty("Mass Excess")
        self.df.rename(columns={"Mass Excess": "Mass Excess (MeV)"}, inplace=True)

        self._convert_column("T1/2 (s)", float)

        if "Abundance (%)" in self.keys():
            self._convert_column_uncertainty("Abundance (%)")

        if "Branching (%)" in self.keys():
            self._convert_column(
                "Branching (%)", lambda x: _parse_float_uncertainty(x, "")
            )

        if "Radiation Energy" in self.keys():
            self._convert_column_uncertainty("Radiation Energy")
            self.df.rename(
                columns={"Radiation Energy": "Radiation Energy (keV)"}, inplace=True
            )

        if "Endpoint Energy" in self.keys():
            self._convert_column_uncertainty("Endpoint Energy")
            self.df.rename(
                columns={"Endpoint Energy": "Endpoint Energy (keV)"}, inplace=True
            )

        if "Radiation Intensity (%)" in self.keys():
            self._convert_column_uncertainty("Radiation Intensity (%)")

        if "Dose" in self.keys():
            self._convert_column_uncertainty("Dose")
            self.df.rename(columns={"Dose": "Dose (MeV / Bq / s)"}, inplace=True)

    def _convert_column(self, col, function):
        """Convert column from string to another type."""
        col_new = []
        for x in self[col]:
            if x == "":
                col_new.append(None)
            else:
                col_new.append(function(x))
        self.df[col] = col_new

    def _convert_column_uncertainty(self, col):
        """Combine column and its uncertainty into one column."""
        col_new = []
        for x, dx in zip(self[col], self[col + " Unc."]):
            x2 = _parse_float_uncertainty(x, dx)
            col_new.append(x2)
        self.df[col] = col_new
        del self.df[col + " Unc."]

    def _sort_columns(self):
        """Sort columns."""
        preferred_order = [
            "Z",
            "Element",
            "A",
            "m",
            "M",
            "N",
            "JPi",
            "T1/2",
            "Energy Level (MeV)",
            "Decay Mode",
            "Branching (%)",
            "Radiation",
            "Radiation subtype",
            "Radiation Energy (keV)",
            "Radiation Intensity (%)",
        ]
        new_cols = []
        for col in preferred_order:
            if col in self.keys():
                new_cols.append(col)
        for col in self.keys():
            if col not in new_cols:
                new_cols.append(col)
        self.df = self.df[new_cols]


class _NuclearWalletCardQuery(_NNDCQuery):
    """NNDC Nuclear Wallet Card data query.

    Nuclear Wallet Card Search can be performed at this URL:
        http://www.nndc.bnl.gov/nudat2/indx_sigma.jsp

    Help page: http://www.nndc.bnl.gov/nudat2/help/wchelp.jsp

      * Energy: Level energy in MeV.
      * JPi: Level spin and parity.
      * Mass Exc: Level Mass Excess in MeV.
      * T1/2 (txt): Level half-life in the format value+units+uncertainty.
      * T1/2 (seconds): value of the level half-life in seconds.
        Levels that are stable are assigned an "infinity" value.
      * Abund.: Natural abundance.
      * Dec Mode: Decay Mode name.
      * Branching (%): Percentual branching ratio for the corresponding
            decay mode.

    Args:
      perform: a boolean dictating whether to immediately perform the query.
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds
      elevel_range : (tuple of float) : range of nuc. energy level (MeV)
      decay : (str) : isotope decay mode from WALLET_DECAY_MODE
      j :  (str) : nuclear spin
      parity : (str) : nuclear parity

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    _URL = "https://www.nndc.bnl.gov/nudat2/sigma_searchi.jsp"
    _DATA = dict(_NNDCQuery._DATA)
    _DATA.update(
        {
            "eled": "disabled",  # E(level) condition on/off
            "elmin": "0",  # E(level) min
            "elmax": "40",  # E(level) max
            "jled": "disabled",  # J_pi(level) condition on/off
            "jlv": "",  # J
            "plv": "ANY",  # parity
            "ord": "zalt",  # order file by Z, A, E(level), T1/2
        }
    )
    _ALLOWED_KEYWORDS = list(_NNDCQuery._ALLOWED_KEYWORDS)
    _ALLOWED_KEYWORDS.extend(["elevel_range", "decay", "j", "parity"])
    _DUMMY_TEXT = """
<html>
<body>

<pre>
A  	Element	Z  	N  	Energy  	JPi           	Mass Exc  	Unc  	T1/2 (txt)               	T1/2 (seconds)           	Abund.    	Unc     	Dec Mode	Branching (%)

 To save this output into a local file, clik on "File" in your browser menu and select "Save as"

</pre></body></html>
"""

    def update(self, **kwargs):
        """Update the search criteria."""
        super().update(**kwargs)
        # handle decay mode
        if "decay" in kwargs:
            if kwargs["decay"].lower() not in WALLET_DECAY_MODE:
                raise NNDCInputError(
                    "Decay mode must be one of {}, not {}".format(
                        WALLET_DECAY_MODE.keys(), kwargs["decay"].lower()
                    )
                )
            warnings.warn(
                'query kwarg "decay" may not be working on NNDC, '
                + "and the user is advised to check the "
                + '"Decay Mode" column of the resulting DataFrame'
            )
            self._data["dmed"] = "enabled"
            self._data["dmn"] = WALLET_DECAY_MODE[kwargs["decay"].lower()]
        # handle energy level condition
        if "elevel_range" in kwargs:
            self._data["eled"] = "enabled"
            self._data["elmin"], self._data["elmax"] = _format_range(
                kwargs["elevel_range"]
            )
            if self._data["elmax"] == "":
                self._data["elmax"] = "1000000000"
        # handle spin and parity
        if "j" in kwargs:
            self._data["jled"] = "enabled"
            self._data["jlv"] = kwargs["j"]
        if "parity" in kwargs:
            if kwargs["parity"].lower() not in PARITIES:
                raise NNDCInputError(
                    "Parity must be one of {}, not {}".format(
                        PARITIES, kwargs["parity"].lower()
                    )
                )
            self._data["jled"] = "enabled"
            self._data["plv"] = kwargs["parity"].upper()


def fetch_wallet_card(**kwargs):
    """Perform NNDC Nuclear Wallet Card data query and return a DataFrame.

    Nuclear Wallet Card Search can be performed at this URL:
        http://www.nndc.bnl.gov/nudat2/indx_sigma.jsp

    Help page: http://www.nndc.bnl.gov/nudat2/help/wchelp.jsp

      * Energy: Level energy in MeV.
      * JPi: Level spin and parity.
      * Mass Exc: Level Mass Excess in MeV.
      * T1/2 (txt): Level half-life in the format value+units+uncertainty.
      * T1/2 (seconds): value of the level half-life in seconds.
        Levels that are stable are assigned an "infinity" value.
      * Abund.: Natural abundance.
      * Dec Mode: Decay Mode name.
      * Branching (%): Percentual branching ratio for the corresponding
            decay mode.

    Args:
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds
      elevel_range : (tuple of float) : range of nuc. energy level (MeV)
      decay : (str) : isotope decay mode from WALLET_DECAY_MODE
      j :  (str) : nuclear spin
      parity : (str) : nuclear parity

    Returns:
      pandas DataFrame with the requested data.

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    query = _NuclearWalletCardQuery(**kwargs)
    return query.df


class _DecayRadiationQuery(_NNDCQuery):
    """NNDC Decay Radiation data query.

    Decay Radiation Search can be performed at this URL:
        http://www.nndc.bnl.gov/nudat2/indx_dec.jsp

    Help page: http://www.nndc.bnl.gov/nudat2/help/dehelp.jsp

      * Radiation: Radiation type, i.e. G for gamma, E for electron.
      * Rad subtype: Further classification of the radiation type.
      * Rad Ene.: Radiation energy in keV.
      * EP Ene.: Beta-decay end point energy in keV.
      * Rad Int.: Radiation absolute intensity.
      * Dose: Radiation dose in MeV/Bq-s
      * Unc: Uncertainties

    Args:
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds
      decay : (str) : isotope decay mode from DECAYRAD_DECAY_MODE
      elevel_range : (tuple of float) : range of parent energy level (MeV)
      type :  (str) : radiation type from DECAYRAD_RADIATION_TYPE
      e_range : (tuple of float) : radiation energy range (keV)
      i_range : (tuple of float): intensity range (percent)

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    _URL = "https://www.nndc.bnl.gov/nudat2/dec_searchi.jsp"
    _DATA = dict(_NNDCQuery._DATA)
    _DATA.update(
        {
            "rted": "enabled",  # radiation type condition on/off
            "rtn": "ANY",  # radiation type: 'ANY' = any, 'G' = gamma
            "reed": "disabled",  # radiation energy condition on/off
            "remin": "0",  # radiation energy min (keV)
            "remax": "10000",  # radiation energy max (keV)
            "ried": "disabled",  # radiation intensity condition on/off
            "rimin": "0",  # radiation intensity min (%)
            "rimax": "100",  # radiation intensity max (%)
            "ord": "zate",  # order file by Z, A, T1/2, E
        }
    )
    _ALLOWED_KEYWORDS = list(_NNDCQuery._ALLOWED_KEYWORDS)
    _ALLOWED_KEYWORDS.extend(["elevel_range", "decay", "type", "e_range", "i_range"])
    _DUMMY_TEXT = """
<html>
<body>
<pre>
A  	Element	Z  	N  	Par. Elevel	Unc. 	JPi       	Dec Mode	T1/2 (txt)    	T1/2 (num)        	Daughter	Radiation	Rad subtype 	Rad Ene.  	Unc       	EP Ene.   	Unc       	Rad Int.  	Unc       	Dose        	Unc

</pre>
To save this output into a local File, clik on "File" in your browser menu and select "Save as"
</body></html>
"""

    def update(self, **kwargs):
        """Update the search criteria."""
        super().update(**kwargs)
        # handle decay mode
        if "decay" in kwargs:
            if kwargs["decay"].lower() not in DECAYRAD_DECAY_MODE:
                raise NNDCInputError(
                    "Decay mode must be one of {}, not {}".format(
                        DECAYRAD_DECAY_MODE.keys(), kwargs["decay"].lower()
                    )
                )
            self._data["dmed"] = "enabled"
            self._data["dmn"] = DECAYRAD_DECAY_MODE[kwargs["decay"].lower()]
        # handle radiation type
        if "type" in kwargs:
            if kwargs["type"].lower() not in DECAYRAD_RADIATION_TYPE:
                raise NNDCInputError(
                    "Radiation type must be one of {}, not {}".format(
                        DECAYRAD_RADIATION_TYPE.keys(), kwargs["type"].lower()
                    )
                )
            self._data["rted"] = "enabled"
            self._data["rtn"] = DECAYRAD_RADIATION_TYPE[kwargs["type"].lower()]
        # handle energy level condition
        self.elevel_range = (0, 1e9)
        if "elevel_range" in kwargs:
            x = _format_range(kwargs["elevel_range"])
            try:
                x0 = float(x[0])
            except ValueError:
                x0 = 0.0
            try:
                x1 = float(x[1])
            except ValueError:
                x1 = 1e9
            self.elevel_range = (x0, x1)
        # handle radiation energy range
        if "e_range" in kwargs:
            self._data["reed"] = "enabled"
            self._data["remin"], self._data["remax"] = _format_range(kwargs["e_range"])
        # handle radiation intensity range
        if "i_range" in kwargs:
            self._data["ried"] = "enabled"
            self._data["rimin"], self._data["rimax"] = _format_range(kwargs["i_range"])


def fetch_decay_radiation(**kwargs):
    """Perform NNDC Decay Radiation data query and return a DataFrame.

    Decay Radiation Search can be performed at this URL:
        http://www.nndc.bnl.gov/nudat2/indx_dec.jsp

    Help page: http://www.nndc.bnl.gov/nudat2/help/dehelp.jsp

      * Radiation: Radiation type, i.e. G for gamma, E for electron.
      * Rad subtype: Further classification of the radiation type.
      * Rad Ene.: Radiation energy in keV.
      * EP Ene.: Beta-decay end point energy in keV.
      * Rad Int.: Radiation absolute intensity.
      * Dose: Radiation dose in MeV/Bq-s
      * Unc: Uncertainties

    Args:
      nuc     : (str) : the name of the isotope (e.g., 'Co-60')
      z, a, n : (int) : Z, A, N of the isotope
      z_range, etc. : (tuple of int) : range of Z, A, or N
      z_any, etc. : (bool) : whether any Z, A, or N is considered
      z_odd, etc. : (bool) : only odd Z, A, or N
      z_even, etc.: (bool) : only even Z, A, or N
      t_range : (tuple of float) : range of isotope half-lives in seconds
      elevel_range : (tuple of float) : range of parent energy level (MeV)
      decay : (str) : isotope decay mode from DECAYRAD_DECAY_MODE
      type :  (str) : radiation type from DECAYRAD_RADIATION_TYPE
      e_range : (tuple of float) : radiation energy range (keV)
      i_range : (tuple of float): intensity range (percent)

    Returns:
      pandas DataFrame with the requested data.

    Raises:
      NNDCInputError: if there is a problem with the input.
      NNDCRequestError: if there was a problem with the data requested.

    """

    query = _DecayRadiationQuery(**kwargs)
    # apply elevel_range filter (hack around the web API)
    elevel = query.df["Energy Level (MeV)"]
    keep = (elevel >= query.elevel_range[0]) & (elevel <= query.elevel_range[1])
    query.df = query.df[keep]
    return query.df



"""Read in a Canberra CNF file.

The CNF format specification appears to be proprietary, so this code is
based on the CNF parsing code in the xylib project:
    https://github.com/wojdyr/xylib

The relevant code is in these files:
    https://github.com/wojdyr/xylib/blob/master/xylib/util.cpp
    https://github.com/wojdyr/xylib/blob/master/xylib/canberra_cnf.cpp
"""

import datetime
import os
import struct
import numpy as np
from ..core import calibration
from .parsers import BecquerelParserError


def _from_little_endian(data, index, n_bytes):
    """Convert bytes starting from index from little endian to an integer."""
    return sum(data[index + j] << 8 * j for j in range(n_bytes))


def _convert_date(data, index):
    """Convert 64-bit number starting at index into a date."""
    # 'data' is the number of 100ns intervals since 17 November 1858, which
    # is the start of the Modified Julian Date (MJD) epoch.
    # 3506716800 is the number of seconds between the start of the MJD epoch
    # to the start of the Unix epoch on 1 January 1970.
    d = _from_little_endian(data, index, 8)
    t = (d / 10000000.0) - 3506716800
    return datetime.datetime.utcfromtimestamp(t)


def _convert_time(data, index):
    """Convert 64-bit number starting at index into a time."""
    # 2^(64) - number of 100ns intervals since timing started
    d = _from_little_endian(data, index, 8)
    d = (pow(2, 64) - 1) & ~d
    return d * 1.0e-7


def _from_pdp11(data, index):
    """Convert 32-bit floating point in DEC PDP-11 format to a double.

    For a description of the format see:
    http://home.kpn.nl/jhm.bonten/computers/bitsandbytes/wordsizes/hidbit.htm
    """
    if (data[index + 1] & 0x80) == 0:
        sign = 1
    else:
        sign = -1
    exb = ((data[index + 1] & 0x7F) << 1) + ((data[index] & 0x80) >> 7)
    if exb == 0:
        if sign == -1:
            return np.NaN
        else:
            return 0.0
    h = (
        data[index + 2] / 256.0 / 256.0 / 256.0
        + data[index + 3] / 256.0 / 256.0
        + (128 + (data[index] & 0x7F)) / 256.0
    )
    return sign * h * pow(2.0, exb - 128.0)


def _read_energy_calibration(data, index):
    """Read the four energy calibration coefficients."""
    coeff = [0.0, 0.0, 0.0, 0.0]
    for i in range(4):
        coeff[i] = _from_pdp11(data, index + 2 * 4 + 28 + 4 * i)
    if coeff[1] == 0.0:
        return None
    return coeff


def read(filename, verbose=False):
    """Parse the CNF file and return a dictionary of data.

    Parameters
    ----------
    filename : str
        The filename of the CNF file to read.
    verbose : bool (optional)
        Whether to print out debugging information. By default False.

    Returns
    -------
    data : dict
        Dictionary of data that can be used to instantiate a Spectrum.
    cal : Calibration
        Energy calibration stored in the file.
    """
    print("Reading CNF file " + filename)
    _, ext = os.path.splitext(filename)
    if ext.lower() != ".cnf":
        raise BecquerelParserError("File extension is incorrect: " + ext)

    # read all of the file into memory
    file_bytes = []
    with open(filename, "rb") as f:
        byte = f.read(1)
        while byte:
            byte_int = struct.unpack("1B", byte)
            file_bytes.append(byte_int[0])
            byte = f.read(1)
    if verbose:
        print("file size (bytes):", len(file_bytes))

    # initialize a dictionary of spectrum data to populate as we parse
    data = {}

    # step through the stored data and parse it
    # skip first 112 bytes of file
    i = 112
    # scan for offsets
    offset_acq = 0
    offset_sam = 0
    offset_eff = 0
    offset_enc = 0
    offset_chan = 0
    for i in range(112, 128 * 1024, 48):
        offset = _from_little_endian(file_bytes, i + 10, 4)
        if (
            (file_bytes[i + 1] == 0x20 and file_bytes[i + 2] == 0x01)
            or (file_bytes[i + 1] == 0)
            or (file_bytes[i + 2] == 0)
        ):
            if file_bytes[i] == 0:
                if offset_acq == 0:
                    offset_acq = offset
                else:
                    offset_enc = offset
            elif file_bytes[i] == 1:
                if offset_sam == 0:
                    offset_sam = offset
            elif file_bytes[i] == 2:
                if offset_eff == 0:
                    offset_eff = offset
            elif file_bytes[i] == 5:
                if offset_chan == 0:
                    offset_chan = offset
            else:
                pass
            if (
                offset_acq != 0
                and offset_sam != 0
                and offset_eff != 0
                and offset_chan != 0
            ):
                break
    if offset_enc == 0:
        offset_enc = offset_acq

    # extract sample information
    if (
        (offset_sam + 48 + 80) >= len(file_bytes)
        or file_bytes[offset_sam] != 1
        or file_bytes[offset_sam + 1] != 0x20
    ):
        if verbose:
            print(offset_sam + 48 + 80, len(file_bytes))
            print(file_bytes[offset_sam], file_bytes[offset_sam + 1])
        raise BecquerelParserError("Sample information not found")
    else:
        sample_name = ""
        for j in range(offset_sam + 48, offset_sam + 48 + 64):
            sample_name += chr(file_bytes[j])
        if verbose:
            print("sample name: ", sample_name)
        sample_id = ""
        for j in range(offset_sam + 112, offset_sam + 112 + 64):
            sample_id += chr(file_bytes[j])
        if verbose:
            print("sample id:   ", sample_id)
        sample_type = ""
        for j in range(offset_sam + 176, offset_sam + 176 + 16):
            sample_type += chr(file_bytes[j])
        if verbose:
            print("sample type: ", sample_type)
        sample_unit = ""
        for j in range(offset_sam + 192, offset_sam + 192 + 64):
            sample_unit += chr(file_bytes[j])
        if verbose:
            print("sample unit: ", sample_unit)
        user_name = ""
        for j in range(offset_sam + 0x02D6, offset_sam + 0x02D6 + 32):
            user_name += chr(file_bytes[j])
        if verbose:
            print("user name:   ", user_name)
        sample_desc = ""
        for j in range(offset_sam + 0x036E, offset_sam + 0x036E + 256):
            sample_desc += chr(file_bytes[j])
        if verbose:
            print("sample desc: ", sample_desc)
        data["sample_name"] = sample_name
        data["sample_id"] = sample_id
        data["sample_type"] = sample_type
        data["sample_unit"] = sample_unit
        data["user_name"] = user_name
        data["sample_description"] = sample_desc

    # extract acquisition information
    if (
        (offset_acq + 48 + 128 + 10 + 4) >= len(file_bytes)
        or file_bytes[offset_acq] != 0
        or file_bytes[offset_acq + 1] != 0x20
    ):
        if verbose:
            print(offset_acq + 48 + 128 + 10 + 4, len(file_bytes))
            print(file_bytes[offset_acq], file_bytes[offset_acq + 1])
        raise BecquerelParserError("Acquisition information not found")
    else:
        offset1 = _from_little_endian(file_bytes, offset_acq + 34, 2)
        offset2 = _from_little_endian(file_bytes, offset_acq + 36, 2)
        offset_pha = offset_acq + 48 + 128
        if (
            chr(file_bytes[offset_pha + 0]) != "P"
            and chr(file_bytes[offset_pha + 1]) != "H"
            and chr(file_bytes[offset_pha + 2]) != "A"
        ):
            raise BecquerelParserError("PHA keyword not found")
        num_channels = 256 * _from_little_endian(file_bytes, offset_pha + 10, 2)
        if num_channels < 256 or num_channels > 16384:
            raise BecquerelParserError("Unexpected number of channels: ", num_channels)
        if verbose:
            print("Number of channels: ", num_channels)

    # extract date and time information
    offset_date = offset_acq + 48 + offset2 + 1
    if offset_date + 24 >= len(file_bytes):
        raise BecquerelParserError("Problem with date offset")
    collection_start = _convert_date(file_bytes, offset_date)
    realtime = _convert_time(file_bytes, offset_date + 8)
    if verbose:
        print("realtime: ", realtime)
    livetime = _convert_time(file_bytes, offset_date + 16)
    if verbose:
        print("livetime: ", livetime)
        print(f"{collection_start:%Y-%m-%d %H:%M:%S}")

    # check data and time information
    if realtime <= 0.0:
        raise BecquerelParserError(f"Realtime not parsed correctly: {realtime}")
    if livetime <= 0.0:
        raise BecquerelParserError(f"Livetime not parsed correctly: {livetime}")
    if livetime > realtime:
        raise BecquerelParserError(f"Livetime > realtime: {livetime} > {realtime}")

    # extract energy calibration information
    offset_cal = offset_enc + 48 + 32 + offset1
    if offset_cal >= len(file_bytes):
        raise BecquerelParserError("Problem with energy calibration offset")
    cal_coeff = _read_energy_calibration(file_bytes, offset_cal)
    if verbose:
        print("calibration coefficients:", cal_coeff)
    if cal_coeff is None:
        if verbose:
            print("Energy calibration - second try")
        cal_coeff = _read_energy_calibration(file_bytes, offset_cal - offset1)
        if verbose:
            print("calibration coefficients:", cal_coeff)
    if cal_coeff is None:
        raise BecquerelParserError("Energy calibration not found")

    # extract channel count data
    if (
        offset_chan + 512 + 4 * num_channels > len(file_bytes)
        or file_bytes[offset_chan] != 5
        or file_bytes[offset_chan + 1] != 0x20
    ):
        raise BecquerelParserError("Channel data not found")
    channels = np.array([], dtype=float)
    counts = np.array([], dtype=float)
    for i in range(0, 2):
        y = _from_little_endian(file_bytes, offset_chan + 512 + 4 * i, 4)
        if y == int(realtime) or y == int(livetime):
            y = 0
        counts = np.append(counts, y)
        channels = np.append(channels, i)
    for i in range(2, num_channels):
        y = _from_little_endian(file_bytes, offset_chan + 512 + 4 * i, 4)
        counts = np.append(counts, y)
        channels = np.append(channels, i)

    # finish populating data dict
    data["realtime"] = realtime
    data["livetime"] = livetime
    data["start_time"] = collection_start
    data["counts"] = counts

    # clean up null characters in any strings
    for key in data.keys():
        if isinstance(data[key], str):
            data[key] = data[key].replace("\x00", " ")
            data[key] = data[key].replace("\x01", " ")
            data[key] = data[key].strip()

    # create an energy calibration object
    cal = calibration.Calibration.from_polynomial(cal_coeff)

    return data, cal



"""Read in an Ortec SPE file."""

import datetime
import os
import warnings
import dateutil.parser
import numpy as np
from ..core import calibration
from .parsers import BecquerelParserError, BecquerelParserWarning


def read(filename, verbose=False):
    """Parse the ASCII SPE file and return a dictionary of data.

    ORTEC's SPE file format is given on page 73 of this document:
        http://www.ortec-online.com/download/ortec-software-file-structure-manual.pdf

    Parameters
    ----------
    filename : str
        The filename of the CNF file to read.
    verbose : bool (optional)
        Whether to print out debugging information. By default False.

    Returns
    -------
    data : dict
        Dictionary of data that can be used to instantiate a Spectrum.
    cal : Calibration
        Energy calibration stored in the file.
    """
    print("SpeFile: Reading file " + filename)
    _, ext = os.path.splitext(filename)
    if ext.lower() != ".spe":
        raise BecquerelParserError("File extension is incorrect: " + ext)

    # initialize a dictionary of spectrum data to populate as we parse
    data = {}

    # parse the file line-by-line
    collection_start = None
    livetime = None
    realtime = None
    counts = []
    channels = []
    cal_coeff = []
    with open(filename) as f:
        # read & remove newlines from end of each line
        lines = [line.strip() for line in f.readlines()]
        i = 0
        while i < len(lines):
            # check whether we have reached a keyword and parse accordingly
            if lines[i] == "$DATE_MEA:":
                i += 1
                collection_start = dateutil.parser.parse(lines[i])
                if verbose:
                    print(collection_start)
            elif lines[i] == "$MEAS_TIM:":
                i += 1
                livetime = float(lines[i].split(" ")[0])
                realtime = float(lines[i].split(" ")[1])
                if verbose:
                    print(livetime, realtime)
            elif lines[i] == "$DATA:":
                i += 1
                first_channel = int(lines[i].split(" ")[0])
                # I don't know why it would be nonzero
                if first_channel != 0:
                    raise BecquerelParserError(
                        f"First channel is not 0: {first_channel}"
                    )
                num_channels = int(lines[i].split(" ")[1])
                if verbose:
                    print(first_channel, num_channels)
                j = first_channel
                while j <= num_channels + first_channel:
                    i += 1
                    counts = np.append(counts, int(lines[i]))
                    channels = np.append(channels, j)
                    j += 1
            elif lines[i] == "$MCA_CAL:":
                i += 1
                n_coeff = int(lines[i])
                i += 1
                for j in range(n_coeff):
                    cal_coeff.append(float(lines[i].split(" ")[j]))
                if verbose:
                    print(cal_coeff)
            elif lines[i].startswith("$"):
                key = lines[i][1:].rstrip(":")
                i += 1
                values = []
                while i < len(lines) and not lines[i].startswith("$"):
                    values.append(lines[i])
                    i += 1
                if i < len(lines):
                    if lines[i].startswith("$"):
                        i -= 1
                if len(values) == 1:
                    values = values[0]
                data[key] = values
            else:
                warnings.warn(
                    f"Line {i + 1} unknown: " + lines[i],
                    BecquerelParserWarning,
                )
            i += 1

    # check the data that were read
    if collection_start is None:
        raise BecquerelParserError("Start time not found")
    if livetime is None:
        raise BecquerelParserError("Live time not found")
    if realtime is None:
        raise BecquerelParserError("Real time not found")

    if realtime <= 0.0:
        raise BecquerelParserError(f"Real time not parsed correctly: {realtime}")
    if livetime <= 0.0:
        raise BecquerelParserError(f"Live time not parsed correctly: {livetime}")
    if livetime > realtime:
        raise BecquerelParserError(f"Livetime > realtime: {livetime} > {realtime}")

    # finish populating data dict
    data["realtime"] = realtime
    data["livetime"] = livetime
    data["start_time"] = collection_start
    data["counts"] = counts

    # create an energy calibration object
    cal = None
    if len(cal_coeff) > 0 and not np.allclose(cal_coeff, 0):
        cal = calibration.Calibration.from_polynomial(cal_coeff)

    return data, cal
